from __future__ import annotations

from collections.abc import Generator, Iterable, MutableSet
from functools import partial
from typing import Any, TypeVar

from jax.tree_util import register_pytree_with_keys
from rich.tree import Tree

from .annotations import PyTree
from .display.batch_dimensions import BatchDimensionIterator, BatchDimensions
from .display.display_generic import display_class, display_generic

__all__ = ['register_graph_as_jax_pytree', 'register_graph_as_nnx_node']


def graph_arrow(directed: bool) -> str:  # noqa: FBT001
    return '⟶' if directed else '↔'


try:
    import networkx as nx
except ImportError:
    msg = "NetworkX not available"
    def register_graph_as_jax_pytree(graph_type: type[Any]) -> None:
        raise RuntimeError(msg)
    def register_graph_as_nnx_node(graph_type: type[Any]) -> None:
        raise RuntimeError(msg)
else:
    T = TypeVar('T', bound="nx.Graph[Any]")

    def validate_node_name(name: Any,
                           arrow: str
                           ) -> None:
        if not isinstance(name, str):
            raise TypeError
        if arrow in name:
            raise ValueError

    def flatten_helper(graph: nx.Graph[Any],
                       arrow: str
                       ) -> Generator[tuple[str, Any], None, None]:
        for name, data in sorted(graph.nodes.data()):
            validate_node_name(name, arrow)
            yield name, data

        def undirected_edge_key(source_target_data: tuple[str, str, Any]) -> tuple[str, ...]:
            source, target, _ = source_target_data
            return tuple(sorted([source, target]))
        edge_data = graph.edges.data()
        directed = isinstance(graph, nx.DiGraph)
        if not directed:
            edge_data = sorted(edge_data, key=undirected_edge_key)
        for source, target, data in edge_data:
            validate_node_name(source, arrow)
            validate_node_name(target, arrow)
            new_source, new_target = ((source, target)
                                      if directed
                                      else sorted([source, target]))
            yield f"{new_source}{arrow}{new_target}", data

    def flatten_graph(graph: nx.Graph[Any],
                      arrow: str
                      ) -> tuple[tuple[tuple[str, Any], ...], None]:
        t = tuple(flatten_helper(graph, arrow))
        return t, None

    def init_graph(graph: nx.Graph[Any],
                   items: Iterable[tuple[str, Any]],
                   arrow: str
                   ) -> None:
        for key, value in items:
            if not isinstance(value, dict):
                raise TypeError
            if arrow in key:
                source, target = key.split(arrow, 2)
                graph.add_edge(source, target, **value)
            else:
                graph.add_node(key, **value)

    def register_graph_as_jax_pytree(graph_type: type[nx.Graph[Any]]) -> None:  # pyright: ignore
        arrow = graph_arrow(issubclass(graph_type, nx.DiGraph))

        def unflatten_tree(names: tuple[str, ...], values: Iterable[Any], /) -> nx.Graph[Any]:
            graph = graph_type()
            init_graph(graph, zip(names, values, strict=True), arrow)
            return graph

        def flatten_with_keys(graph: nx.Graph[Any], /
                              ) -> tuple[Iterable[tuple[str, Any]], tuple[str, ...]]:
            names_and_values = tuple(flatten_helper(graph, arrow))
            names = tuple(name for name, _ in names_and_values)
            return (names_and_values, names)

        def flatten_tree(graph: nx.Graph[Any], /) -> tuple[Iterable[PyTree], tuple[str, ...]]:
            names_and_values = tuple(flatten_helper(graph, arrow))
            names, values = zip(*names_and_values, strict=True)
            return values, names

        register_pytree_with_keys(graph_type, flatten_with_keys, unflatten_tree, flatten_tree)

    @display_generic.register(nx.Graph)
    def _(value: nx.Graph[Any],
          *,
          seen: MutableSet[int] | None = None,
          show_values: bool = True,
          key: str = '',
          batch_dims: BatchDimensions | None = None) -> Tree:
        if seen is None:
            seen = set()
        arrow = graph_arrow(isinstance(value, nx.DiGraph))
        retval = display_class(key, type(value))
        bdi = BatchDimensionIterator(batch_dims)
        for name, node in value.nodes.items():
            sub_batch_dims = bdi.advance(node)
            retval.children.append(display_generic(node, seen=seen, show_values=show_values,
                                                   key=name, batch_dims=sub_batch_dims))
        for (source, target), edge in value.edges.items():
            key = f"{source}{arrow}{target}"
            sub_batch_dims = bdi.advance(edge)
            retval.children.append(display_generic(edge, seen=seen, show_values=show_values,
                                                   key=key, batch_dims=sub_batch_dims))
        return retval

    try:
        from flax.experimental.nnx.nnx.graph_utils import register_node_type
    except ImportError:
        msg = "NNX not available"
        def register_graph_as_nnx_node(graph_type: type[Any]) -> None:
            raise RuntimeError(msg)
    else:
        def register_graph_as_nnx_node(graph_type: type[T]) -> None:  # pyright: ignore
            arrow = graph_arrow(issubclass(graph_type, nx.DiGraph))

            def get_key_graph(graph: nx.Graph[Any], key: str) -> Any:
                if arrow in key:
                    source, target = key.split(arrow, 2)
                    return graph.edges[source, target]
                return graph.nodes[key]

            def set_key_graph(graph: nx.Graph[Any], key: str, value: Any) -> nx.Graph[Any]:
                if not isinstance(value, dict):
                    raise TypeError
                d: dict[str, Any]
                if arrow in key:
                    source, target = key.split(arrow, 2)
                    d = graph.edges[source, target]
                else:
                    d = graph.nodes[key]
                d.clear()
                d.update(value)
                return graph

            def has_key_graph(graph: nx.Graph[Any], key: str) -> bool:
                if arrow in key:
                    source, target = key.split(arrow, 2)
                    # Care here to return true for only ordered edges.
                    return (source, target) in list(graph.edges)
                return key in graph

            def all_keys_helper(graph: nx.Graph[Any]
                                ) -> Generator[str, None, None]:
                for name in graph.nodes:
                    validate_node_name(name, arrow)
                    yield name
                for source, target in graph.edges:
                    validate_node_name(source, arrow)
                    validate_node_name(target, arrow)
                    yield f"{source}{arrow}{target}"

            def all_keys_graph(graph: nx.Graph[Any]) -> tuple[str, ...]:
                return tuple(all_keys_helper(graph))

            def create_empty_graph(metadata: None) -> T:
                return graph_type()

            register_node_type(graph_type,
                               partial(flatten_graph, arrow=arrow),
                               get_key_graph,
                               set_key_graph,
                               has_key_graph,
                               all_keys_graph,
                               create_empty=create_empty_graph,
                               init=partial(init_graph, arrow=arrow))



from __future__ import annotations

from functools import partial, singledispatch
from math import prod
from numbers import Complex, Integral, Real
from operator import and_
from typing import Any

import jax
import jax.numpy as jnp
import numpy as np
from jax.tree_util import tree_flatten, tree_map, tree_reduce

from .annotations import Array, PyTree
from .dtypes import default_tols

__all__ = ['assert_tree_allclose', 'get_relative_test_string', 'get_test_string', 'tree_allclose']


def assert_tree_allclose(actual: PyTree,
                         desired: PyTree,
                         original_name: str | None = None,
                         original_value: PyTree | None = None,
                         *,
                         rtol: float | None = None,
                         atol: float | None = None) -> None:
    """Assert that an actual pytree matches a desired pytree.

    If the assertion fails, a passing test string is printed::

    ```python
    from tjax import assert_tree_allclose, RealNumeric
    from tjax.dataclasses import dataclass

    @dataclass
    class A:
        x: RealNumeric
        y: RealNumeric

    @dataclass
    class B:
        z: A

    original = B(A(1.2, 3.4))
    desired = B(A(3.0, 4.0))
    actual = B(A(1.2, 5.2))

    assert_tree_allclose(actual, desired)
    ```
    This prints::
    ```
    Tree leaves don't match at position 0 with rtol=0.0001 and atol=1e-06.
    Mismatched elements: 1 / 1 (100%)
    Maximum absolute difference: 1.8
    Maximum relative difference: 0.6

    Actual: B(z=A(x=1.2, y=5.2))
    Desired: B(z=A(x=3.0, y=4.0))
    Test string:
    desired = B(A(1.2, 5.2))
    ```
    The test string can then be pasted.

    Args:
        actual: The actual value.
        desired: The desired value.
        original_name: The variable name that contains the original value.
        original_value: The original value.  This is usually a pytree like a dataclass that has the
            same type as actual and desired, but contains different values.
        rtol: The relative tolerance of the comparisons in the assertion.
        atol: The absolute tolerance of the comparisons in the assertion.
    """
    flattened_actual, structure_actual = tree_flatten(actual)
    flattened_desired, structure_desired = tree_flatten(desired)
    if structure_actual != structure_desired:
        msg = f"\nTree structure mismatch.\nActual: {actual}\nDesired: {desired}\n"
        raise AssertionError(msg)

    for i, (actual_, desired_) in enumerate(zip(flattened_actual, flattened_desired, strict=True)):
        dtype = jnp.result_type(actual_, desired_)
        tols = default_tols(dtype.type, rtol=rtol, atol=atol)
        try:
            np.testing.assert_allclose(actual_, desired_, **tols)
        except AssertionError as exception:
            old_message = exception.args[0].split('\n')
            best_part_of_old_message = "\n".join(old_message[3:6]).replace("Max ", "Maximum ")
            test_string = (get_relative_test_string(actual, original_name, original_value, **tols)
                           if original_name is not None and original_value is not None
                           else get_test_string(actual, **tols))
            test_string = "desired = " + test_string
            # style_config = yapf.style.CreatePEP8Style()
            # style_config['COLUMN_LIMIT'] = column_limit
            # test_string = yapf.yapf_api.FormatCode(test_string, style_config=style_config)[0]
            message = (
                f"\nTree leaves don't match at position {i} with rtol={tols['rtol']} and "
                f"atol={tols['atol']}.\n"
                f"{best_part_of_old_message}\n\n"
                f"Actual: {actual}\nDesired: {desired}\n"
                f"Test string:\n{test_string}")
            raise AssertionError(message) from None


def tree_allclose(actual: PyTree,
                  desired: PyTree,
                  rtol: float | None = None,
                  atol: float | None = None) -> bool:
    """Return whether two pytrees are close.

    Args:
        actual: The actual value.
        desired: The desired value.
        rtol: The relative tolerance of the comparisons in the comparison.
        atol: The absolute tolerance of the comparisons in the comparison.
    """
    def allclose(actual_array: Array, desired_array: Array) -> bool:
        dtype = jnp.result_type(actual_array, desired_array)
        tols = default_tols(dtype.type, rtol=rtol, atol=atol)
        return bool(jnp.allclose(actual_array, desired_array, **tols))

    return tree_reduce(and_, tree_map(allclose, actual, desired), True)


# get test string ----------------------------------------------------------------------------------
# Redefinition typing errors in this file are due to https://github.com/python/mypy/issues/2904.
@singledispatch
def get_test_string(actual: Any, rtol: float, atol: float) -> str:
    """Produce a short string of Python code that produces the actual value.

    Args:
        actual: The actual value that was produced.
        rtol: The relative tolerance of the comparisons in the assertion.
        atol: The absolute tolerance of the comparisons in the assertion.
    """
    return repr(actual)


@get_test_string.register(np.ndarray)
@get_test_string.register(jax.Array)
def _(actual: Array | jax.Array, rtol: float, atol: float) -> str:
    if prod(actual.shape) == 0:
        return f"np.empty({actual.shape}, dtype=np.{actual.dtype})"
    with np.printoptions(formatter={'float_kind': partial(_inexact_number_to_string, rtol=rtol,
                                                          atol=atol),
                                    'complex_kind': partial(_inexact_number_to_string, rtol=rtol,
                                                            atol=atol)}):
        return "np." + repr(np.asarray(actual)).replace(' ]', ']').replace(' ,', ',').replace(
            '  ', ' ').replace('dtype=', 'dtype=np.')


@get_test_string.register
def _(actual: Complex, rtol: float, atol: float) -> str:
    x = float(actual) if isinstance(actual, Real) else complex(actual)
    return _inexact_number_to_string(x, rtol, atol)


@get_test_string.register
def _(actual: Integral, rtol: float, atol: float) -> str:
    return str(actual)


@get_test_string.register(list)
@get_test_string.register(tuple)
def _(actual: list[Any] | tuple[Any], rtol: float, atol: float) -> str:
    is_list = isinstance(actual, list)
    is_named_tuple = not is_list and type(actual).__name__ != 'tuple'
    return ((type(actual).__name__ if is_named_tuple else "")
            + ("[" if is_list else "(")
            + ", ".join(get_test_string(sub_actual, rtol, atol)
                        for i, sub_actual in enumerate(actual))
            + (',' if len(actual) == 1 else '')
            + ("]" if is_list else ")"))


@get_test_string.register(dict)
def _(actual: dict[Any, Any], rtol: float, atol: float) -> str:
    return '{' + ",\n".join(repr(key) + ': ' + get_test_string(sub_actual, rtol, atol)
                            for key, sub_actual in actual.items()) + '}'


# get relative test string -------------------------------------------------------------------------
@singledispatch
def get_relative_test_string(actual: Any,
                             original_name: str,
                             original: Any,
                             rtol: float,
                             atol: float) -> str:
    """Produce code for use in tests based on actual and produced values.

    Args:
        original_name: The name of the variable containing an original value.
        actual: The actual value that was produced, and that should be the desired value.
        original: The original value.
        rtol: The relative tolerance of the comparisons in the assertion.
        atol: The absolute tolerance of the comparisons in the assertion.

    Returns:
        A string of Python code that produces the desired value from an "original value" (could be
        zeroed-out, for example).
    """
    return str(actual)


@get_relative_test_string.register(np.ndarray)
@get_relative_test_string.register(jax.Array)
def _(actual: Array | jax.Array, original_name: str, original: Any, rtol: float,
      atol: float) -> str:
    with np.printoptions(formatter={'float_kind': partial(_inexact_number_to_string, rtol=rtol,
                                                          atol=atol),
                                    'complex_kind': partial(_inexact_number_to_string, rtol=rtol,
                                                            atol=atol)}):
        return "np." + repr(np.asarray(actual)).replace(' ]', ']').replace(' ,', ',').replace('  ',
                                                                                              ' ')


@get_relative_test_string.register
def _(actual: Complex, original_name: str, original: Any, rtol: float, atol: float) -> str:
    return _inexact_number_to_string(actual, rtol, atol)  # type: ignore[arg-type] # pyright: ignore


@get_relative_test_string.register
def _(actual: Integral, original_name: str, original: Any, rtol: float, atol: float) -> str:
    return str(actual)


@get_relative_test_string.register(list)
@get_relative_test_string.register(tuple)
def _(actual: list[Any] | tuple[Any], original_name: str, original: Any, rtol: float,
      atol: float) -> str:
    is_list = isinstance(actual, list)
    return (("[" if is_list else "(")
            + ", ".join(get_relative_test_string(f"{original_name}[{i}]",
                                                 sub_actual, sub_original,
                                                 rtol, atol)
                        for i, (sub_actual, sub_original) in enumerate(zip(actual, original,
                                                                           strict=True)))
            + ("]" if is_list else ")"))


@get_relative_test_string.register(dict)
def _(actual: dict[Any, Any], original_name: str, original: Any, rtol: float, atol: float) -> str:
    if not isinstance(original, dict):
        raise TypeError

    def relative_string(key: Any, sub_actual: Any) -> str:
        return get_relative_test_string(
            f"{original_name}[{key}]", sub_actual, original[key], rtol, atol)

    return '{' + ",\n".join(repr(key) + ': ' + relative_string(key, sub_actual)
                            for key, sub_actual in actual.items()) + '}'


# Private functions --------------------------------------------------------------------------------
def _float_to_string_with_precision(x: complex, precision: int) -> str:
    with np.printoptions(precision=precision, floatmode='maxprec'):
        return repr(np.asarray(x))[6:-1]


def _inexact_number_to_string(x: complex | np.inexact[Any], rtol: float, atol: float) -> str:
    y: float | complex
    if isinstance(x, Real):  # type: ignore[unreachable]
        y = float(x)  # type: ignore[unreachable]
    elif isinstance(x, Complex):  # type: ignore[unreachable]
        y = complex(x)  # type: ignore[unreachable]
    else:
        raise TypeError
    retval = ""  # type: ignore[unreachable]
    for i in range(20):
        retval = _float_to_string_with_precision(y, i)
        if np.isclose(complex(retval), x, rtol=rtol, atol=atol):
            break
    return retval



from __future__ import annotations

from collections.abc import Callable
from typing import Any, TypeVar

__all__ = ['tree_map_with_path']


T = TypeVar('T')
def tree_map_with_path(structure: Any,
                       transform: Callable[[T, tuple[str, ...]], T]
                       ) -> Any:
    def f(structure: Any,
          transform: Callable[[T, tuple[str, ...]], T],
          path: tuple[str, ...]
          ) -> Any:
        if isinstance(structure, dict):
            out_structure = {}
            for key, value in structure.items():
                out_structure[key] = f(value, transform, (*path, key))
            return out_structure
        return transform(structure, path)
    return f(structure, transform, ())



from ._src.dataclasses.dataclass import DataclassInstance, TDataclassInstance, dataclass
from ._src.dataclasses.flax import DataClassModule, module_field
from ._src.dataclasses.helpers import as_shallow_dict, field

__all__ = [
    'DataClassModule',
    'DataclassInstance',
    'TDataclassInstance',
    'as_shallow_dict',
    'dataclass',
    'field',
    'module_field']



from __future__ import annotations

from typing import Any, Generic, TypeVar

import jax.numpy as jnp
from jax import vjp
from jax.tree_util import tree_map
from typing_extensions import override

from tjax.dataclasses import dataclass

from ..annotations import PyTree
from ..shims import custom_vjp_method
from .augmented import State
from .base import Parameters, Trajectory
from .comparing import ComparingIteratedFunction, ComparingState
from .iterated_function import Comparand, IteratedFunction, TheAugmentedState

__all__ = ['ComparingIteratedFunctionWithCombinator', 'IteratedFunctionWithCombinator']


Differentiand = TypeVar('Differentiand', bound=PyTree)


@dataclass
class _ZResiduals(Generic[Parameters, State, Comparand, Differentiand, TheAugmentedState]):
    outer_iterated_function: IteratedFunctionWithCombinator[Parameters, State, Comparand,
                                                            Differentiand, Any, TheAugmentedState]
    outer_theta: Parameters
    x_star: State


@dataclass
class _ZParameters(Generic[Parameters, State, Differentiand]):
    outer_theta: Parameters
    x_star: State
    x_star_differentiand: Differentiand
    x_star_bar_differentiand: Differentiand


def _ffp_bwd(residuals: _ZResiduals[Parameters, State, Comparand, Differentiand, TheAugmentedState],
             augmented_star_bar: TheAugmentedState) -> tuple[None, Parameters, None]:
    """The backward pass of the fixed point finder.

    Args:
        residuals: residuals produced by _ffp_fwd.
        augmented_star_bar: cotangents
    Returns:
        theta_bar: cotangents for theta
        zeroed_xs: cotangents for initial_state
    """
    outer_iterated_function = residuals.outer_iterated_function
    outer_theta = residuals.outer_theta
    x_star = residuals.x_star
    x_star_differentiand = outer_iterated_function.extract_differentiand(outer_theta, x_star)
    x_star_bar = augmented_star_bar.current_state
    x_star_bar_differentiand = outer_iterated_function.extract_differentiand(outer_theta,
                                                                             x_star_bar)

    def f_of_theta(some_theta: Parameters) -> Differentiand:
        state = outer_iterated_function.expected_state(some_theta, x_star)
        return outer_iterated_function.extract_differentiand(outer_theta, state)

    z_iterator = _ZIterate(minimum_iterations=outer_iterated_function.z_minimum_iterations,
                           maximum_iterations=outer_iterated_function.z_maximum_iterations,
                           rtol=outer_iterated_function.rtol, atol=outer_iterated_function.atol,
                           z_minimum_iterations=outer_iterated_function.z_minimum_iterations,
                           z_maximum_iterations=outer_iterated_function.z_maximum_iterations,
                           iterated_function=outer_iterated_function)
    z_parameters = _ZParameters(residuals.outer_theta, x_star, x_star_differentiand,
                                x_star_bar_differentiand)
    augmented = z_iterator.find_fixed_point(z_parameters, x_star_bar_differentiand)
    z_star_differentiand: Differentiand = augmented.current_state

    _, df_by_dtheta = vjp(f_of_theta, residuals.outer_theta)
    theta_bar, = df_by_dtheta(z_star_differentiand)
    return None, theta_bar, None


@dataclass
class IteratedFunctionWithCombinator(
        IteratedFunction[Parameters, State, Comparand, Trajectory, TheAugmentedState],
        Generic[Parameters, State, Comparand, Differentiand, Trajectory, TheAugmentedState]):
    """An IteratedFunctionWithCombinator is an IteratedFunction that invokes a combinator.

    It allows differentiation works through the fixed point.  Besides inheriting from this class, no
    other action is necessary to get this capability.

    It is a generic class with all of the parameters of IteratedFunction, and Differentiand, which
    is the type of the *portion of the state* with respect to which derivatives at the fixed point
    are calculated.

    Attributes:
        z_maximum_iterations:
            The maximum number of iterations to use to evaluate the adjoint's fixed point.
    """
    z_minimum_iterations: int
    z_maximum_iterations: int

    # Overridden methods ---------------------------------------------------------------------------
    @custom_vjp_method
    @override
    def find_fixed_point(self,  # type: ignore[override] # pyright: ignore
                         theta: Parameters,
                         initial_state: State) -> TheAugmentedState:
        """Find the fixed point.

        This is a separate method because it has a bound custom VJP.

        Args:
            theta: The parameters for which gradients can be calculated.
            initial_state: An initial guess of the final state.
        Returns: The augmented state at the fixed point.
        """
        return super().find_fixed_point(theta, initial_state)

    # Abstract methods -----------------------------------------------------------------------------
    def extract_differentiand(self, theta: Parameters, state: State) -> Differentiand:
        """Extract the differentiable values in a state.

        It is used by the combinator to find cotangents.

        Returns: The differentiable values in the state.
        """
        raise NotImplementedError

    def implant_differentiand(self,
                              theta: Parameters,
                              state: State,
                              differentiand: Differentiand) -> State:
        """Implant the differentiand into the state.

        Args:
            theta: The parameters for which gradients can be calculated.
            state: A state that will provide nondifferentiable values.
            differentiand: A differentiand that will provide differentiable values.
        Returns: A state containing differentiable from the differentiand and nondifferentiable
            values from the inputted state.
        """
        raise NotImplementedError

    # Apply vjp ------------------------------------------------------------------------------------
    def _ffp_fwd(self,
                 theta: Parameters,
                 initial_state: State) -> tuple[TheAugmentedState,
                                                _ZResiduals[Parameters, State, Comparand,
                                                            Differentiand, TheAugmentedState]]:
        """The forward pass of the fixed point finder.

        Args:
            theta: The parameters for which gradients can be calculated.
            initial_state: An initial guess of the final state.

        Returns:
            x_star: the result of the minimization.
            residuals: residuals used in _ffp_bwd.
        """
        augmented: TheAugmentedState = self.find_fixed_point(theta, initial_state)
        return augmented, _ZResiduals(self, theta, augmented.current_state)

    find_fixed_point.defvjp(_ffp_fwd, _ffp_bwd)


class ComparingIteratedFunctionWithCombinator(
        IteratedFunctionWithCombinator[Parameters, State, Comparand, Differentiand, Trajectory,
                                       ComparingState[State, Comparand]],
        ComparingIteratedFunction[Parameters, State, Comparand, Trajectory],
        Generic[Parameters, State, Comparand, Differentiand, Trajectory]):
    pass


@dataclass
class _ZIterate(ComparingIteratedFunctionWithCombinator[
        _ZParameters[Parameters, State, Differentiand],
        Differentiand,
        Differentiand,
        Differentiand,
        None],
        Generic[Parameters, State, Comparand, TheAugmentedState, Differentiand]):
    """The state of _ZIterate is the differentiand of the outer iterated function."""
    iterated_function: IteratedFunctionWithCombinator[
        Parameters, State, Comparand, Differentiand, Any, TheAugmentedState]

    # Implemented methods --------------------------------------------------------------------------
    @override
    def expected_state(self,
                       theta: _ZParameters[Parameters, State, Differentiand],
                       state: Differentiand) -> Differentiand:
        return self.sampled_state(theta, state)

    @override
    def sampled_state(self,
                      theta: _ZParameters[Parameters, State, Differentiand],
                      state: Differentiand) -> Differentiand:
        # The state should be called z, but we can't change the interface because of Liskov's
        # substitution principle.
        z = state
        del state

        def f_of_x(x_differentiand: Differentiand) -> Differentiand:
            x = self.iterated_function.implant_differentiand(theta.outer_theta, theta.x_star,
                                                             x_differentiand)
            state = self.iterated_function.expected_state(theta.outer_theta, x)
            return self.iterated_function.extract_differentiand(theta.outer_theta, state)

        _, df_by_dx = vjp(f_of_x, theta.x_star_differentiand)
        df_by_dx_times_z, = df_by_dx(z)
        return tree_map(jnp.add, theta.x_star_bar_differentiand, df_by_dx_times_z)

    @override
    def extract_comparand(self, state: Differentiand) -> Differentiand:
        return state

    @override
    def extract_differentiand(self,
                              theta: _ZParameters[Parameters, State, Differentiand],
                              state: Differentiand) -> Differentiand:
        return state

    @override
    def implant_differentiand(self,
                              theta: _ZParameters[Parameters, State, Differentiand],
                              state: Differentiand,
                              differentiand: Differentiand) -> Differentiand:
        return differentiand



from __future__ import annotations

import dataclasses
from collections.abc import Callable, Hashable, Iterable
from typing import Any, ClassVar, Protocol, cast, overload, runtime_checkable

from jax.tree_util import register_pytree_with_keys
from typing_extensions import dataclass_transform

from ..annotations import PyTree
from ..testing import get_relative_test_string, get_test_string, tree_allclose
from .helpers import field

__all__ = ['DataclassInstance', 'TDataclassInstance', 'dataclass']


@runtime_checkable
class DataclassInstance(Protocol):
    __dataclass_fields__: ClassVar[dict[str, dataclasses.Field[Any]]]


@runtime_checkable
class TDataclassInstance(DataclassInstance, Protocol):
    static_fields: ClassVar[list[str]]
    dynamic_fields: ClassVar[list[str]]


@overload
@dataclass_transform(frozen_default=True, field_specifiers=(field,))
def dataclass(*, init: bool = True, repr: bool = True, eq: bool = True,  # noqa: A002
              order: bool = False) -> Callable[[type[Any]], type[TDataclassInstance]]:
    ...


@overload
@dataclass_transform(frozen_default=True, field_specifiers=(field,))
def dataclass(cls: type[Any], /, *, init: bool = True, repr: bool = True,  # noqa: A002
              eq: bool = True, order: bool = False) -> type[TDataclassInstance]:
    ...


@dataclass_transform(frozen_default=True, field_specifiers=(field,))
def dataclass(cls: type[Any] | None = None, /, *, init: bool = True, repr: bool = True,  # noqa: A002
              eq: bool = True, order: bool = False
              ) -> type[TDataclassInstance] | Callable[[type[Any]], type[TDataclassInstance]]:
    """A dataclass creator that creates a pytree.

    Returns the same class as was passed in, with dunder methods added based on the fields defined
    in the class.

    Examines PEP 526 annotations to determine fields.  Default values for fields are provided using
    assignment.

    To mark fields as static fields rather than JAX pytree fields, use the `field` function.
    In JAX, a static attribute is one that induces recompilation of a function when it changes, and
    consequently there is more flexibility about what can be done with such an attribute.

    For example::
    ```python
    from __future__ import annotations

    from typing import ClassVar

    import jax.numpy as jnp
    from tjax import Array, dataclass
    from tjax.dataclasses import field
    from jax import grad

    @dataclass
    class LearnedParameter:
        weight: Array
        constrain_positive: bool = field(static=True)
        minimum_positive_weight: ClassVar[Array] = 1e-6

        def trained(self,
                    self_bar: LearnedParameter,
                    learning_rate: float) -> LearnedParameter:
            weight_bar = self_bar.weight
            weight = self.weight - weight_bar * learning_rate
            if self.constrain_positive:
                weight = jnp.maximum(weight, self.minimum_positive_weight)
            return LearnedParameter(weight=weight,
                                    constrain_positive=self.constrain_positive)

    def loss(w: LearnedParameter) -> float:
        return jnp.square(w.weight - 3.3)

    w = LearnedParameter(2.0, True)
    w_bar = grad(loss)(w)
    new_w = w.trained(w_bar, 1e-4)
    ```

    Since this dataclass is a pytree, all of JAX's functions that accept pytrees work with it,
    including iteration, differentiation, and `jax.tree_util` functions.

    Another benefit is the display of dataclasses.  `print(new_w)` gives::
    ```
    LearnedParameter
        weight=Jax Array ()
                2.0003
        constrain_positive=True
    ```
    """
    if cls is None:
        def f(x: type[Any], /) -> type[TDataclassInstance]:
            return dataclass(x, init=init, repr=repr, eq=eq, order=order)
        return f  # Type checking support partial is poor.
    non_none_cls = cls

    # Apply dataclass function to cls.
    data_clz: type[TDataclassInstance] = cast(type[TDataclassInstance],
                                              dataclasses.dataclass(init=init, repr=repr, eq=eq,
                                                                    order=order, frozen=True)(cls))

    # Partition fields into static, and dynamic; and assign these to the class.
    static_fields: list[str] = []
    dynamic_fields: list[str] = []
    for field_info in dataclasses.fields(data_clz):
        if field_info.metadata.get('pytree_node', True):
            dynamic_fields.append(field_info.name)
        else:
            static_fields.append(field_info.name)
    data_clz.static_fields = static_fields
    data_clz.dynamic_fields = dynamic_fields

    # Register the class as a Jax PyTree.
    def flatten_with_keys(x: Any, /) -> tuple[Iterable[tuple[str, PyTree]], Hashable]:
        hashed = tuple(getattr(x, name) for name in static_fields)
        trees = tuple((name, getattr(x, name)) for name in dynamic_fields)
        return trees, hashed

    def unflatten(hashed: Hashable, trees: Iterable[PyTree], /) -> Any:
        if not isinstance(hashed, tuple):
            raise TypeError
        hashed_args = dict(zip(static_fields, hashed, strict=True))
        tree_args = dict(zip(dynamic_fields, trees, strict=True))
        return non_none_cls(**hashed_args, **tree_args)

    def flatten(x: Any, /) -> tuple[Iterable[PyTree], Hashable]:
        hashed = tuple(getattr(x, name) for name in static_fields)
        trees = tuple(getattr(x, name) for name in dynamic_fields)
        return trees, hashed

    register_pytree_with_keys(data_clz, flatten_with_keys, unflatten, flatten)

    # Register the dynamically-dispatched functions.
    get_test_string.register(data_clz, get_dataclass_test_string)
    get_relative_test_string.register(data_clz, get_relative_dataclass_test_string)
    return data_clz


def get_dataclass_test_string(actual: Any, rtol: float, atol: float) -> str:
    retval = f"{type(actual).__name__}("
    retval += ",\n".join(((f"{fn.name}=" if fn.kw_only else "")
                          + get_test_string(getattr(actual, fn.name), rtol, atol))
                         for fn in dataclasses.fields(actual))
    retval += ")"
    return retval


def get_relative_dataclass_test_string(actual: Any,
                                       original_name: str,
                                       original: Any,
                                       rtol: float,
                                       atol: float) -> str:
    retval = f"replace({original_name}, "
    retval += ",\n".join(
        f"{fn}=" + get_relative_test_string(f"{original_name}.{fn}",
                                            getattr(actual, fn),
                                            getattr(original, fn),
                                            rtol,
                                            atol)
        for fn in actual.dynamic_fields
        if not tree_allclose(getattr(actual, fn), getattr(original, fn), rtol=rtol, atol=atol))
    retval += ")"
    return retval



from __future__ import annotations

from dataclasses import KW_ONLY, InitVar, dataclass
from typing import Any, TypeAlias

from flax.experimental import nnx
from typing_extensions import dataclass_transform, override

from .helpers import field

__all__ = ['DataClassModule', 'module_field']


def module_field(*, init: bool = False) -> Any:
    """A field that contains submodules."""
    return field(init=init, default=None, kw_only=True)


NamedChildren: TypeAlias = tuple[tuple[str, nnx.State], tuple[str, nnx.GraphDef[Any]]]
Children: TypeAlias = tuple[nnx.State, nnx.GraphDef[Any]]


@dataclass_transform(field_specifiers=(module_field,
                                       nnx.field,
                                       nnx.treenode_field,
                                       nnx.variable_field,
                                       nnx.param_field))
class _DataClassModule(nnx.Module):
    @override
    def __init_subclass__(cls,
                          *,
                          init: bool = True,
                          repr: bool = True,  # noqa: A002
                          eq: bool = True,
                          order: bool = False,
                          kw_only: bool = False,
                          **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs, experimental_pytree=True)
        dataclass(init=init, repr=repr, eq=eq, order=order, kw_only=kw_only)(cls)


class DataClassModule(_DataClassModule):
    _: KW_ONLY
    rngs: InitVar[nnx.Rngs] = nnx.field()

    def __post_init__(self, rngs: nnx.Rngs) -> None:
        pass



from __future__ import annotations

from collections.abc import Callable
from functools import partial
from typing import Any, TypeVar, cast

import jax.numpy as jnp
from jax import vjp
from jax.tree_util import tree_map, tree_structure

from .annotations import JaxRealArray, RealNumeric
from .display.tapped import tapped_print_generic
from .shims import custom_jvp, custom_vjp

__all__ = [
    'copy_cotangent',
    'cotangent_combinator',
    'print_cotangent',
    'replace_cotangent',
    'scale_cotangent']


X = TypeVar('X')
XT = TypeVar('XT', bound=tuple[Any, ...])
Y = TypeVar('Y')


# scale_cotangent ----------------------------------------------------------------------------------
def scale_cotangent(x: X, scale: RealNumeric) -> X:
    return x


def _scale_cotangent_jvp(scale: RealNumeric, primals: tuple[X], tangents: tuple[X]) -> tuple[X, X]:
    x, = primals
    x_dot, = tangents
    scaled_x_dot = tree_map(lambda x_dot_i: x_dot_i * scale, x_dot)
    return x, scaled_x_dot


scale_cotangent = custom_jvp(scale_cotangent, nondiff_argnums=(1,))
# The type variables don't match, so this type error needs to be ignored.
scale_cotangent.defjvp(_scale_cotangent_jvp)  # pyright: ignore


# copy_cotangent -----------------------------------------------------------------------------------
@custom_vjp
def copy_cotangent(x: X, y: X) -> X:
    assert tree_structure(x) == tree_structure(y)
    return x


def _copy_cotangent_fwd(x: X, y: X) -> tuple[X, None]:
    assert tree_structure(x) == tree_structure(y)
    return x, None


def _copy_cotangent_bwd(residuals: None, x_bar: X) -> tuple[X, X]:
    del residuals
    return x_bar, x_bar


# Pyright can't infer types because custom_vjp doesn't yet depend on ParamSpec.
copy_cotangent.defvjp(_copy_cotangent_fwd, _copy_cotangent_bwd)


# replace_cotangent --------------------------------------------------------------------------------
@custom_vjp
def replace_cotangent(x: X, new_cotangent: X) -> X:
    assert tree_structure(x) == tree_structure(new_cotangent)
    return x


def _replace_cotangent_fwd(x: X, new_cotangent: X) -> tuple[X, X]:
    assert tree_structure(x) == tree_structure(new_cotangent)
    return x, new_cotangent


def _replace_cotangent_bwd(residuals: X, x_bar: X) -> tuple[X, X]:
    assert tree_structure(residuals) == tree_structure(x_bar)
    return residuals, x_bar


# Pyright can't infer types because custom_vjp doesn't yet depend on ParamSpec.
replace_cotangent.defvjp(_replace_cotangent_fwd, _replace_cotangent_bwd)


# print_cotangent ----------------------------------------------------------------------------------
@partial(custom_vjp, static_argnums=(1,))
def print_cotangent(u: X, name: str | None = None) -> X:
    return u


def _print_cotangent_fwd(u: X, name: str | None) -> tuple[X, None]:
    return u, None


def _print_cotangent_bwd(name: str | None, residuals: None, x_bar: X) -> tuple[X]:
    del residuals
    x_bar = (tapped_print_generic(x_bar)
        if name is None
        else tapped_print_generic(
            **{name: x_bar}))  # type: ignore[call-overload] # pyright: ignore
    return (x_bar,)


# https://github.com/python/mypy/issues/14802
print_cotangent.defvjp(_print_cotangent_fwd, _print_cotangent_bwd)


# cotangent_combinator -----------------------------------------------------------------------------
def cotangent_combinator(f: Callable[..., tuple[XT, Y]],
                         args_tuples: tuple[tuple[Any, ...], ...],
                         aux_cotangent_scales: tuple[RealNumeric, ...] | None) -> tuple[XT, Y]:
    """Run a function once, but send differerent cotangents back to each input.

    Args:
        f: A function that accepts positional arguments and returns xs, y where xs is a tuple of
            length n.
        args_tuples: n copies of the same tuple of positional arguments accepted by f.
        aux_cotangent_scales: If provided, scale each of the cotangents.
    Returns: The pair (xs, y).

    The purpose of the cotangent combinator is to take cotangents of each of the elements of x, and
    send them back through to each of the corresponding argument tuples.
    """
    return f(*args_tuples[0])


cotangent_combinator = custom_vjp(cotangent_combinator, static_argnums=(0, 2))


def _cotangent_combinator_fwd(f: Callable[..., tuple[XT, Y]],
                              args_tuples: tuple[tuple[Any, ...], ...],
                              aux_cotangent_scales: tuple[RealNumeric, ...] | None
                              ) -> tuple[tuple[XT, Y],
                                         Callable[[tuple[XT, Y]], tuple[Any, ...]]]:
    return vjp(f, *args_tuples[0])


def _cotangent_combinator_bwd(f: Callable[..., tuple[XT, Y]],
                              aux_cotangent_scales: tuple[RealNumeric, ...] | None,
                              f_vjp: Callable[[tuple[XT, Y]], tuple[Any, ...]],
                              xy_bar: tuple[XT, Y]
                              ) -> tuple[Any, ...]:
    xs_bar, y_bar = xy_bar
    if aux_cotangent_scales is None:
        aux_cotangent_scales = tuple(1.0 for _ in xs_bar)
    xs_zero = tuple(tree_map(jnp.zeros_like, x_bar)
                    for x_bar in xs_bar)
    all_args_bar = []
    for i, (x_bar, aux_cotangent_scale) in enumerate(zip(xs_bar, aux_cotangent_scales,
                                                         strict=True)):
        scaled_y_bar = tree_map(lambda y_bar_i, scale=aux_cotangent_scale: y_bar_i * scale,
                                y_bar)
        this_xs_bar = cast(XT, (xs_zero[:i]
                                + (x_bar,)
                                + xs_zero[i + 1:]))
        this_result_bar = (this_xs_bar, scaled_y_bar)
        this_args_bar = f_vjp(this_result_bar)
        all_args_bar.append(this_args_bar)
    return (tuple(all_args_bar),)


cotangent_combinator.defvjp(_cotangent_combinator_fwd, _cotangent_combinator_bwd)


# augment_cotangent --------------------------------------------------------------------------------
@custom_vjp
def augment_cotangent(loss: JaxRealArray,
                      x: X,
                      cotangent_addend: X
                      ) -> JaxRealArray:
    """Create a cotangent for x equal to the cotangent_addend times the loss cotangent."""
    assert tree_structure(x) == tree_structure(cotangent_addend)
    return loss


# Private functions --------------------------------------------------------------------------------
def _augment_cotangent_fwd(loss: JaxRealArray,
                           x: X,
                           cotangent_addend: X
                           ) -> tuple[JaxRealArray, X]:
    assert tree_structure(x) == tree_structure(cotangent_addend)
    return loss, cotangent_addend


def _augment_cotangent_bwd(cotangent_addend: X,
                           loss_bar: JaxRealArray
                           ) -> tuple[None, X, None]:
    x_bar = tree_map(lambda x: x * loss_bar, cotangent_addend)
    return (None, x_bar, None)


augment_cotangent.defvjp(_augment_cotangent_fwd, _augment_cotangent_bwd)



import numpy as np

from .annotations import NumpyComplexArray, NumpyRealArray

__all__ = ['create_diagonal_array', 'np_abs_square']


def np_abs_square(x: NumpyComplexArray) -> NumpyRealArray:
    return np.square(x.real) + np.square(x.imag)  # pyright: ignore


def create_diagonal_array(m: NumpyRealArray) -> NumpyRealArray:
    """A vectorized version of diagonal.

    Args:
        m: Has shape (*k, n)
    Returns: Array with shape (*k, n, n) and the elements of m on the diagonals.
    """
    indices = (..., *np.diag_indices(m.shape[-1]))
    retval = np.zeros((*m.shape, m.shape[-1]), dtype=m.dtype)
    retval[indices] = m
    return retval



"""Tools for the differential programming library [JAX](https://github.com/google/jax)."""
from . import dataclasses, fixed_point, gradient
from ._src.abstract_method_decorators import JaxAbstractClass, abstract_custom_jvp, abstract_jit
from ._src.annotations import (Array, BooleanArray, BooleanNumeric, Complex, ComplexArray,
                               ComplexNumeric, Integral, IntegralArray, IntegralNumeric, JaxArray,
                               JaxBooleanArray, JaxComplexArray, JaxIntegralArray, JaxRealArray,
                               KeyArray, NumpyArray, NumpyBooleanArray, NumpyBooleanNumeric,
                               NumpyComplexArray, NumpyComplexNumeric, NumpyIntegralArray,
                               NumpyIntegralNumeric, NumpyRealArray, NumpyRealNumeric, PyTree, Real,
                               RealArray, RealNumeric, Shape, ShapeLike, SliceLike,
                               TapFunctionTransforms)
from ._src.cotangent_tools import (copy_cotangent, cotangent_combinator, print_cotangent,
                                   replace_cotangent, scale_cotangent)
from ._src.display.display_generic import display_generic
from ._src.display.print_generic import print_generic
from ._src.display.tapped import tapped_print_generic
from ._src.dtypes import default_atol, default_rtol, default_tols
from ._src.graph import register_graph_as_jax_pytree, register_graph_as_nnx_node
from ._src.leaky_integral import (diffused_leaky_integrate, leaky_covariance, leaky_data_weight,
                                  leaky_integrate, leaky_integrate_time_series)
from ._src.math_tools import (abs_square, divide_nonnegative, divide_where, inverse_softplus,
                              is_scalar, zero_tangent_like)
from ._src.numpy_tools import np_abs_square, create_diagonal_array
from ._src.partial import Partial
from ._src.shims import custom_jvp, custom_jvp_method, custom_vjp, custom_vjp_method, jit
from ._src.testing import (assert_tree_allclose, get_relative_test_string, get_test_string,
                           tree_allclose)
from ._src.tree_tools import tree_map_with_path

__all__ = [
    'Array',
    'BooleanArray',
    'BooleanNumeric',
    'Complex',
    'ComplexArray',
    'ComplexNumeric',
    'Integral',
    'IntegralArray',
    'IntegralNumeric',
    'JaxAbstractClass',
    'JaxArray',
    'JaxBooleanArray',
    'JaxComplexArray',
    'JaxIntegralArray',
    'JaxRealArray',
    'KeyArray',
    'NumpyArray',
    'NumpyBooleanArray',
    'NumpyBooleanNumeric',
    'NumpyComplexArray',
    'NumpyComplexNumeric',
    'NumpyIntegralArray',
    'NumpyIntegralNumeric',
    'NumpyRealArray',
    'NumpyRealNumeric',
    'Partial',
    'PyTree',
    'Real',
    'RealArray',
    'RealNumeric',
    'Shape',
    'ShapeLike',
    'SliceLike',
    'TapFunctionTransforms',
    'abs_square',
    'abstract_custom_jvp',
    'abstract_jit',
    'assert_tree_allclose',
    'copy_cotangent',
    'cotangent_combinator',
    'create_diagonal_array',
    'custom_jvp',
    'custom_jvp_method',
    'custom_vjp',
    'custom_vjp_method',
    'dataclasses',
    'default_atol',
    'default_rtol',
    'default_tols',
    'diffused_leaky_integrate',
    'display_generic',
    'divide_nonnegative',
    'divide_where',
    'fixed_point',
    'get_relative_test_string',
    'get_test_string',
    'gradient',
    'inverse_softplus',
    'is_scalar',
    'jit',
    'leaky_covariance',
    'leaky_data_weight',
    'leaky_integrate',
    'leaky_integrate_time_series',
    'np_abs_square',
    'print_cotangent',
    'print_generic',
    'register_graph_as_jax_pytree',
    'register_graph_as_nnx_node',
    'replace_cotangent',
    'scale_cotangent',
    'tapped_print_generic',
    'tree_allclose',
    'tree_map_with_path',
    'zero_tangent_like']



from tjax._src.fixed_point.augmented import AugmentedState
from tjax._src.fixed_point.base import IteratedFunctionBase
from tjax._src.fixed_point.combinator import (ComparingIteratedFunctionWithCombinator,
                                              IteratedFunctionWithCombinator)
from tjax._src.fixed_point.comparing import ComparingIteratedFunction, ComparingState
from tjax._src.fixed_point.iterated_function import IteratedFunction
from tjax._src.fixed_point.simple_scan import SimpleScan
from tjax._src.fixed_point.stochastic import (StochasticIteratedFunction,
                                              StochasticIteratedFunctionWithCombinator,
                                              StochasticState)
from tjax._src.fixed_point.zero import ZeroIteratedFunction, ZeroIteratedFunctionWithCombinator

__all__ = ['AugmentedState', 'ComparingIteratedFunction', 'ComparingIteratedFunctionWithCombinator',
           'ComparingState', 'IteratedFunction', 'IteratedFunctionBase',
           'IteratedFunctionWithCombinator', 'SimpleScan', 'StochasticIteratedFunction',
           'StochasticIteratedFunctionWithCombinator', 'StochasticState', 'ZeroIteratedFunction',
           'ZeroIteratedFunctionWithCombinator']



from __future__ import annotations

from typing import Any, TypeVar, overload

from jax.experimental.host_callback import id_tap
from jax.tree_util import tree_flatten, tree_unflatten
from rich.console import Console

from ..annotations import TapFunctionTransforms
from .batch_dimensions import BatchDimensions, combine_batch_dimensions
from .print_generic import print_generic

__all__ = ['tapped_print_generic']

_U = TypeVar('_U')


@overload
def tapped_print_generic(*args: Any,
                         raise_on_nan: bool = True,
                         console: Console | None = None,
                         no_jvp: bool = False,
                         result: None = None,
                         **kwargs: Any
                         ) -> Any:
    ...


@overload
def tapped_print_generic(*args: Any,
                         raise_on_nan: bool = True,
                         console: Console | None = None,
                         no_jvp: bool = False,
                         result: _U,
                         **kwargs: Any
                         ) -> _U:
    ...


def tapped_print_generic(*args: Any,
                         raise_on_nan: bool = True,
                         console: Console | None = None,
                         no_jvp: bool = False,
                         result: Any = None,
                         **kwargs: Any
                         ) -> Any:
    """Uses print_generic in a tapped function.

    Args:
        raise_on_nan: Assert if NaN is found in the output.
        console: The console that formats the output.
        no_jvp: Stifle printout of JVP tangents.
        result: A tracer to be returned to ensure sequencing.
        args: Positional arguments to be printed.  Only dynamic arguments are allowed.
        kwargs: Keyword arguments to be printed.  Only static keys and dynamic values are allowed.
    Returns: The value of result, or else the lone element of args and kwargs.
    """
    leaves, tree_def = tree_flatten((args, kwargs))

    def tap(tap_leaves: list[Any],
            transforms: TapFunctionTransforms
            ) -> None:
        args, kwargs = tree_unflatten(tree_def, tap_leaves)
        batch_dims: BatchDimensions = None
        flags = []
        for transform_name, transform_dict in transforms:
            if transform_name == 'batch':
                batch_dims = combine_batch_dimensions(batch_dims, transform_dict['batch_dims'])
                continue
            if no_jvp and transform_name == 'jvp':
                return
            if transform_name in {'jvp', 'mask', 'transpose'}:
                flags.append(transform_name)
                continue
        modified_kwargs = ({key + f" [{', '.join(flags)}]": value
                            for key, value in kwargs.items()}
                           if flags
                           else kwargs)
        print_generic(*args, batch_dims=batch_dims, raise_on_nan=raise_on_nan, console=console,
                      **modified_kwargs)

    if result is None:
        if args:
            assert len(args) == 1
            result = args[0]
        elif kwargs:
            assert len(kwargs) == 1
            result = next(iter(kwargs.values()))
        else:
            assert False  # noqa: PT015
    return id_tap(tap, leaves, result=result)  # type: ignore[no-untyped-call]



from __future__ import annotations

from typing import Any, TypeAlias

from jax.tree_util import tree_leaves
from typing_extensions import override

__all__ = ['BatchDimensionIterator', 'BatchDimensions']


_SimpleBatchDimensions: TypeAlias = tuple[int | None, ...] | None
BatchDimensions: TypeAlias = tuple[tuple[int, ...] | None, ...] | None


def combine_batch_dimensions(x: BatchDimensions, y: _SimpleBatchDimensions) -> BatchDimensions:
    if y is None:
        return x
    # Convert y to BatchDimensions.
    y_bd = tuple(None if yi is None else (yi,)
                 for yi in y)
    if x is None:
        return y_bd
    # Merge x and y.
    assert len(x) == len(y)
    return tuple(yi if xi is None
                 else xi if yi is None
                 else xi + yi
                 for xi, yi in zip(x, y_bd, strict=True))


class BatchDimensionIterator:
    """An iterator for batch dimensions.

    Batch dimensions is created by using jax.vmap on a function.  V-mapped dimensions can have
    non-unity batch dimensions.  This class keeps track of these dimensions and provides
    iterator-like behavior for displaying the batch dimensions along corresponding elements.

    Args:
        batch_dims: A tuple of integers corresponding to the leaves of the PyTree.  This class
        keeps track of these dimensions.
    """
    @override
    def __init__(self, batch_dims: BatchDimensions | None = None):
        super().__init__()
        self.batch_dims = batch_dims
        self.i = 0

    def advance(self, value: Any) -> BatchDimensions:
        """Advance the iterator.

        Args:
            value: The next sub-element of the PyTree.  It need not be a leaf.
        Returns: The batch dimensions of the leaves of value.
        """
        if self.batch_dims is None:
            return None
        n = len(tree_leaves(value))
        old_i = self.i
        self.i += n
        return self.batch_dims[old_i: self.i]

    def check_done(self) -> None:
        """Asserts that all of the batch dimensions are accounted for."""
        if self.batch_dims is not None:
            assert self.i == len(self.batch_dims)



from __future__ import annotations

from functools import partial
from typing import Any, Generic, TypeVar

import jax.numpy as jnp
from jax import Array
from jax.lax import while_loop

from tjax.dataclasses import dataclass

from ..annotations import BooleanNumeric, IntegralNumeric, PyTree, RealNumeric
from .augmented import AugmentedState, State
from .base import IteratedFunctionBase, Parameters, Trajectory

__all__ = ['IteratedFunction']


Comparand = TypeVar('Comparand', bound=PyTree)
TheAugmentedState = TypeVar('TheAugmentedState', bound=AugmentedState[Any])


@dataclass
class IteratedFunction(IteratedFunctionBase[Parameters, State, Trajectory, TheAugmentedState],
                       Generic[Parameters, State, Comparand, Trajectory, TheAugmentedState]):
    """An IteratedFunction object models an iterated function.

    It is a generic class in terms of five generic types, all of which are pytrees:
        * Parameters, which models the iteration parameters,
        * State, which is the state at each iteration,
        * Comparand, which is calculated from state and compared between iterations to check
          convergence,
        * Trajectory, which is the type produced by sample_trajectory, and
        * TheAugmentedState, which is typically defined by one of the tjax subclasses.

    The two main methods of IteratedFunction are
    * find_fixed_point, which iterates until the fixed point is reached or other stopping conditions
      are met, and
    * sample_trajectory, which iterates for a fixed number of iterations and returns a trajectory.

    Args:
        minimum_iterations: The minimum number of iterations for this fixed point solver.  This must
            be positive.
        maximum_iterations: The maximum number of iterations for this fixed point solver.  This must
            be positive.
        rtol: The relative tolerance for the comparison stopping condition.
        atol: The absolute tolerance for the comparison stopping condition.
    """
    minimum_iterations: IntegralNumeric
    maximum_iterations: IntegralNumeric
    rtol: RealNumeric
    atol: RealNumeric

    # New methods ----------------------------------------------------------------------------------
    def find_fixed_point(self,
                         theta: Parameters,
                         initial_state: State) -> TheAugmentedState:
        """The fixed point finder.

        Args:
            theta: The parameters for which gradients can be calculated.
            initial_state: An initial guess of the final state.

        Returns:
            state: The augmented state at the fixed point.
        """
        def f(augmented: TheAugmentedState) -> TheAugmentedState:
            new_state = self.sampled_state(theta, augmented.current_state)
            return self.iterate_augmented(new_state, augmented)

        return while_loop(partial(self.state_needs_iteration, theta),
                          f,
                          self.initial_augmented(initial_state))

    def debug_fixed_point(self, theta: Parameters, initial_state: State) -> TheAugmentedState:
        """This method is identical to find_fixed_point, but avoids using while_loop."""
        augmented = self.initial_augmented(initial_state)
        while self.state_needs_iteration(theta, augmented):
            new_state = self.sampled_state(theta, augmented.current_state)
            augmented = self.iterate_augmented(new_state, augmented)
        return augmented

    def state_needs_iteration(self, theta: Parameters, augmented: TheAugmentedState) -> Array:
        """Whether the state needs to be iterated.

        Args:
            theta: The parameters.
            augmented: The state.
        Returns: True while iteration needs to continue.
        """
        enough_iterations = augmented.iterations >= self.minimum_iterations
        converged = self.converged(augmented)
        not_too_many_iterations = augmented.iterations < self.maximum_iterations
        not_converged_or_done = jnp.logical_not(
            jnp.logical_and(enough_iterations, converged))
        return jnp.logical_and(not_too_many_iterations,
                               not_converged_or_done)

    # Abstract methods -----------------------------------------------------------------------------
    def expected_state(self, theta: Parameters, state: State) -> State:
        """The expected value of the next state given the old one.

        This is used by the combinator.
        """
        raise NotImplementedError

    def converged(self, augmented: TheAugmentedState) -> BooleanNumeric:
        """A Boolean Array of shape () indicating whether the pytrees are close."""
        raise NotImplementedError

    def extract_comparand(self, state: State) -> Comparand:
        """Extracts the "comparand" from the state.

        This is a subset of the values in the state that are compared by subclasses to detect
        convergence:
        * In ZeroIteratedFunction, the comparand is compared with zero.
        * In ComparingIteratedFunction, the comparand is compared in successive states.
        * In StochasticIteratedFunction, the mean and variance of the comparand are compared.
        """
        raise NotImplementedError

    def minimum_tolerances(self, augmented: TheAugmentedState) -> tuple[RealNumeric, RealNumeric]:
        """The minimum tolerances that would lead to convergence now.

        Returns:
            The minimum value of atol that would lead to convergence now.
            The minimum value of rtol that would lead to convergence now.
        """
        raise NotImplementedError



from __future__ import annotations

import math
from collections.abc import Iterable, Mapping, MutableSet
from dataclasses import fields, is_dataclass
from functools import singledispatch
from numbers import Number
from types import FunctionType
from typing import Any

import numpy as np
from jax import Array
from jax.errors import TracerArrayConversionError
from jax.interpreters.batching import BatchTracer
from jax.tree_util import PyTreeDef
from jaxlib.xla_extension import PjitFunction
from rich.table import Table
from rich.text import Text
from rich.tree import Tree

from tjax.dataclasses import DataclassInstance

from ..annotations import NumpyArray
from .batch_dimensions import BatchDimensionIterator, BatchDimensions
from .colors import solarized

__all__ = ['display_class', 'display_generic']


# Constants ----------------------------------------------------------------------------------------
# Numeric (warm)
_numpy_array_color = solarized['magenta']
_jax_array_color = solarized['orange']
_number_color = solarized['yellow']
# solarized['red']
# Classes (cool)
_class_color = solarized['violet']
# solarized['cyan']
_type_color = solarized['green']
_string_color = solarized['base0']
# Other
_key_color = solarized['blue']
_separator_color = solarized['base00']
_table_color = solarized['base02']
_unknown_color = f"{solarized['red']} bold"
_seen_color = solarized['red']


# Extra imports ------------------------------------------------------------------------------------
FlaxModule: type[Any]
FlaxVariable: type[Any]
try:
    from flax.experimental import nnx
except ImportError:
    flax_loaded = False
    def is_node_type(x: type[Any]) -> bool:
        return False
    FlaxModule = type(None)
    FlaxVariable = type(None)
else:
    flax_loaded = True
    FlaxModule = nnx.Module
    FlaxVariable = nnx.Variable
    is_node_type = nnx.graph_utils.is_node_type


@singledispatch
def display_generic(value: Any,
                    *,
                    seen: MutableSet[int] | None = None,
                    show_values: bool = True,
                    key: str = '',
                    batch_dims: BatchDimensions | None = None) -> Tree:
    if seen is None:
        seen = set()
    if (x := _verify(value, seen, key)) is not None:
        return x
    if is_dataclass(value) and not isinstance(value, type):
        return _display_dataclass(value, seen=seen, show_values=show_values, key=key,
                                  batch_dims=batch_dims)
    return _display_object(value, seen=seen, show_values=show_values, key=key,
                           batch_dims=batch_dims)
    # _assemble(key, Text(str(value), style=_unknown_color))


@display_generic.register
def _(value: str,
      *,
      seen: MutableSet[int] | None = None,
      show_values: bool = True,
      key: str = '',
      batch_dims: tuple[int | None, ...] | None = None) -> Tree:
    if seen is None:
        seen = set()
    if (x := _verify(value, seen, key)) is not None:
        return x
    return _assemble(key, Text(f'"{value}"', style=_string_color))


@display_generic.register(type)
def _(value: type[Any],
      *,
      seen: MutableSet[int] | None = None,
      show_values: bool = True,
      key: str = '',
      batch_dims: BatchDimensions | None = None) -> Tree:
    if seen is None:
        seen = set()
    if (x := _verify(value, seen, key)) is not None:
        return x
    return _assemble(key, Text(f"type[{value.__name__}]", style=_type_color))


@display_generic.register(np.ndarray)
def _(value: NumpyArray,
      *,
      seen: MutableSet[int] | None = None,
      show_values: bool = True,
      key: str = '',
      batch_dims: BatchDimensions | None = None) -> Tree:
    if seen is None:
        seen = set()
    if (x := _verify(value, seen, key)) is not None:
        return x
    extracted_batch_sizes, shape = _batched_axis_sizes_from_array_and_dims(value, batch_dims)
    batch_str = _batch_str(extracted_batch_sizes)
    retval = _assemble(key,
                       Text(f"NumPy Array {shape} {value.dtype}{batch_str}",
                            style=_numpy_array_color))
    _show_array(retval, value)
    return retval


@display_generic.register
def _(value: Array,
      *,
      seen: MutableSet[int] | None = None,
      show_values: bool = True,
      key: str = '',
      batch_dims: BatchDimensions | None = None) -> Tree:
    if seen is None:
        seen = set()
    if (x := _verify(value, seen, key)) is not None:
        return x
    # extracted_batch_sizes = _batched_axis_sizes_from_jax_array(value)
    extracted_batch_sizes = ()
    batch_str = _batch_str(extracted_batch_sizes)
    retval = _assemble(key,
                       Text(f"Jax Array {value.shape} {value.dtype}{batch_str}",
                            style=_jax_array_color))
    try:
        np_value = np.asarray(value)
    except TracerArrayConversionError:
        pass
    else:
        _show_array(retval, np_value)
    return retval


@display_generic.register(type(None))
@display_generic.register(Number)
def _(value: None | Number,
      *,
      seen: MutableSet[int] | None = None,
      show_values: bool = True,
      key: str = '',
      batch_dims: BatchDimensions | None = None) -> Tree:
    if seen is None:
        seen = set()
    if (x := _verify(value, seen, key)) is not None:
        return x
    return _assemble(key, Text(str(value), style=_number_color))


@display_generic.register(Mapping)
def _(value: Mapping[Any, Any],
      *,
      seen: MutableSet[int] | None = None,
      show_values: bool = True,
      key: str = '',
      batch_dims: BatchDimensions | None = None) -> Tree:
    if seen is None:
        seen = set()
    if (x := _verify(value, seen, key)) is not None:
        return x
    retval = display_class(key, type(value))
    for sub_batch_dims, (sub_key, sub_value) in zip(_batch_dimension_iterator(value.values(),
                                                                              batch_dims),
                                                    value.items(),
                                                    strict=True):
        retval.children.append(display_generic(sub_value, seen=seen, show_values=show_values,
                                               key=str(sub_key), batch_dims=sub_batch_dims))
    return retval


@display_generic.register(tuple)
@display_generic.register(list)
def _(value: tuple[Any, ...] | list[Any],
      *,
      seen: MutableSet[int] | None = None,
      show_values: bool = True,
      key: str = '',
      batch_dims: BatchDimensions | None = None) -> Tree:
    if seen is None:
        seen = set()
    if (x := _verify(value, seen, key)) is not None:
        return x
    retval = display_class(key, type(value))
    for sub_batch_dims, sub_value in zip(_batch_dimension_iterator(value, batch_dims), value,
                                         strict=True):
        retval.children.append(display_generic(sub_value, seen=seen, show_values=show_values,
                                               key="", batch_dims=sub_batch_dims))
    return retval


@display_generic.register(PyTreeDef)
def _(value: PyTreeDef,
      *,
      seen: MutableSet[int] | None = None,
      show_values: bool = True,
      key: str = '',
      batch_dims: BatchDimensions | None = None) -> Tree:
    if seen is None:
        seen = set()
    if (x := _verify(value, seen, key)) is not None:
        return x
    retval = display_class(key, type(value))
    retval.children.append(display_generic(hash(value), seen=seen, show_values=show_values,
                                           key="hash"))
    return retval


@display_generic.register(FunctionType)
@display_generic.register(PjitFunction)
def _(value: PyTreeDef,
      *,
      seen: MutableSet[int] | None = None,
      show_values: bool = True,
      key: str = '',
      batch_dims: BatchDimensions | None = None) -> Tree:
    if seen is None:
        seen = set()
    if (x := _verify(value, seen, key)) is not None:
        return x
    name = getattr(value, '__qualname__', "")
    retval = display_class(key, type(value))
    retval.children.append(display_generic(name, seen=seen, show_values=show_values, key="name"))
    return retval


if flax_loaded:
    @display_generic.register(FlaxVariable)
    def _(value: nnx.Variable[Any],
          *,
          seen: MutableSet[int] | None = None,
          show_values: bool = True,
          key: str = '',
          batch_dims: BatchDimensions | None = None) -> Tree:
        if seen is None:
            seen = set()
        if (x := _verify(value, seen, key)) is not None:
            return x
        retval = display_class(key, type(value))
        variables = _variables(value)
        variables = {key: sub_value
                     for key, sub_value in variables.items()
                     if not (key.endswith('_hooks') and value)}
        for name, sub_value in variables.items():
            retval.children.append(display_generic(sub_value, seen=seen, show_values=show_values,
                                                   key=name, batch_dims=None))
        return retval


# Public unexported functions ----------------------------------------------------------------------
def display_class(key: str, cls: type[Any]) -> Tree:
    name = cls.__name__
    tags = []
    if is_dataclass(cls):
        tags.append('dataclass')
    if flax_loaded and cls not in {int, float, tuple, list, set, dict}:
        if issubclass(cls, FlaxModule):
            tags.append('flax-module')
        elif is_node_type(cls):
            tags.append('flax-node')
    if tags:
        name += f"[{','.join(tags)}]"
    return _assemble(key, Text(name, style=_class_color))


# Private functions --------------------------------------------------------------------------------
def _display_dataclass(value: DataclassInstance,
                       *,
                       seen: MutableSet[int] | None = None,
                       show_values: bool = True,
                       key: str = '',
                       batch_dims: BatchDimensions | None = None) -> Tree:
    if seen is None:
        seen = set()
    retval = display_class(key, type(value))
    bdi = BatchDimensionIterator(batch_dims)
    names = set()
    for field_info in fields(value):
        name = field_info.name
        names.add(name)
        display_name = name
        if not field_info.init:
            display_name += ' (module)'
        sub_value = getattr(value, name, None)
        sub_batch_dims = (bdi.advance(sub_value)
                          if field_info.metadata.get('pytree_node', True)
                          else None)
        retval.children.append(display_generic(sub_value, seen=seen, show_values=show_values,
                                               key=display_name, batch_dims=sub_batch_dims))
    variables = _variables(value)
    for name, sub_value in variables.items():
        if name in names:
            continue
        retval.children.append(display_generic(sub_value, seen=seen, show_values=show_values,
                                               key=name + '*', batch_dims=None))
    return retval


def _display_object(value: Any,
                    *,
                    seen: MutableSet[int],
                    show_values: bool = True,
                    key: str = '',
                    batch_dims: BatchDimensions | None = None) -> Tree:
    retval = display_class(key, type(value))
    variables = _variables(value)
    for name, sub_value in variables.items():
        retval.children.append(display_generic(sub_value, seen=seen, show_values=show_values,
                                               key=name, batch_dims=None))
    return retval


def _variables(value: Any) -> dict[str, Any]:
    try:
        variables = vars(value)
    except TypeError:
        variables = ({name: getattr(value, name) for name in value.__slots__}
                     if hasattr(value, '__slots__')
                     else {})
    return {key: value
            for key, value in variables.items()
            if key != '_module__state'
            if not key.startswith('__')}


def _verify(value: Any,
            seen: MutableSet[int],
            key: str) -> Tree | None:
    if id(value) in seen:
        return _assemble(key, Text('<seen>', style=_seen_color))
    if is_dataclass(value) and not isinstance(value, type):
        seen.add(id(value))
    return None


def _assemble(key: str,
              type_text: Text,
              separator: str = '=') -> Tree:
    """Returns: A Rich Tree for a given key-value pair."""
    if key:
        return Tree(Text.assemble(Text(key, style=_key_color),
                                  Text(separator, style=_separator_color),
                                  type_text))
    return Tree(type_text)


@singledispatch
def _format_number(x: Any) -> str:
    return str(x)


@_format_number.register(np.inexact)
def _(x: np.inexact[Any]) -> str:
    return f"{x:.4f}"


def _show_array(tree: Tree, array: NumpyArray) -> None:
    """Add a representation of array to the Rich tree."""
    if not issubclass(array.dtype.type, np.bool_ | np.number):
        return
    if math.prod(array.shape) == 0:
        return
    if 1 in array.shape:
        _show_array(tree,
                    array[tuple[int | slice, ...](0 if s == 1 else slice(None)
                                                        for s in array.shape)])
        return
    if any(x > 12 for x in array.shape) or len(array.shape) > 2:  # noqa: PLR2004
        xarray = np.asarray(array)
        tree.children.append(display_generic(float(np.mean(xarray)), seen=set(), key="mean"))
        tree.children.append(display_generic(float(np.std(xarray)), seen=set(), key="deviation"))
        return
    if len(array.shape) == 0:
        tree.add(_format_number(array[()]))
        return
    table = Table(show_header=False,
                  show_edge=False,
                  style=_table_color)
    for _ in range(array.shape[-1]):
        table.add_column()
    if len(array.shape) == 1:
        table.add_row(*(_format_number(array[i])
                        for i in range(array.shape[0])))
    elif len(array.shape) == 2:  # noqa: PLR2004
        for j in range(array.shape[0]):
            table.add_row(*(_format_number(array[j, i])
                            for i in range(array.shape[1])))
    tree.add(table)


def _batch_dimension_iterator(values: Iterable[Any],
                              batch_dims: BatchDimensions | None = None
                              ) -> Iterable[BatchDimensions]:
    """Traverse values and batch_dims in parallel.

    Returns: An iterable of BatchDimensions objects for sub-elements of value.
    """
    bdi = BatchDimensionIterator(batch_dims)
    for value in values:
        yield bdi.advance(value)
    bdi.check_done()


def _batch_str(extracted_batch_sizes: tuple[int, ...]) -> str:
    return (f" batched over axes of size {extracted_batch_sizes}"
            if extracted_batch_sizes
            else "")


def _batched_axis_sizes_from_array_and_dims(value: NumpyArray,
                                            batch_dims: BatchDimensions
                                            ) -> tuple[tuple[int, ...],
                                                       tuple[int, ...]]:
    """Discover the shapes of the batched axes and the shape of the array without them.

    This function uses batch dimension information and a NumPy array, so it is only effective on
    jax.Arrays in tapped display.

    Returns: A tuple:
        The size of the axes that are batched over for a given Numpy array.
        The shape of the Numpy array without the batched axes.
    """
    shape = value.shape
    if batch_dims is None:
        return (), shape
    assert len(batch_dims) == 1
    batch_dim_tuple = batch_dims[0]
    if batch_dim_tuple is None:
        return (), shape
    assert isinstance(batch_dim_tuple, tuple)

    batch_sizes = []
    for batch_dim in reversed(batch_dim_tuple):
        batch_sizes.append(shape[batch_dim])
        shape = shape[: batch_dim] + shape[batch_dim + 1:]
    return tuple(batch_sizes), shape


def _batched_axis_sizes_from_jax_array(value: Array) -> tuple[int, ...]:
    """Discover the shapes of the batched axes.

    This function works by looking for BatchTracer objects, so it is only effective on jax.Arrays in
    non-tapped display.  This no longer works.

    Returns: The size of the axes that are batched over for a given Jax array.
    """
    batch_sizes = []
    while isinstance(value, BatchTracer):
        batch_sizes.append(value.val.shape[value.batch_dim])
        value = value.val
    return tuple(reversed(batch_sizes))



from __future__ import annotations

from functools import partial
from typing import Generic

import jax.numpy as jnp
from jax.tree_util import tree_map, tree_reduce
from typing_extensions import override

from tjax.dataclasses import dataclass

from ..annotations import JaxBooleanArray, JaxRealArray
from ..math_tools import divide_nonnegative
from .augmented import AugmentedState, State
from .iterated_function import Comparand, IteratedFunction, Parameters, Trajectory

__all__ = ['ComparingIteratedFunction', 'ComparingState']


@dataclass
class ComparingState(AugmentedState[State], Generic[State, Comparand]):
    last_state: Comparand


@dataclass
class ComparingIteratedFunction(
        IteratedFunction[Parameters, State, Comparand, Trajectory,
                         ComparingState[State, Comparand]],
        Generic[Parameters, State, Comparand, Trajectory]):
    """An iterated function that converges when part of two successive states are nearly equal."""

    # Implemented methods --------------------------------------------------------------------------
    @override
    def initial_augmented(self, initial_state: State) -> ComparingState[State, Comparand]:
        return ComparingState(current_state=initial_state,
                              iterations=0,
                              last_state=self.extract_comparand(initial_state))

    @override
    def expected_state(self, theta: Parameters, state: State) -> State:
        return self.sampled_state(theta, state)

    @override
    def iterate_augmented(self,
                          new_state: State,
                          augmented: ComparingState[State, Comparand]) -> (
                              ComparingState[State, Comparand]):
        return ComparingState(current_state=new_state,
                              iterations=augmented.iterations + 1,
                              last_state=self.extract_comparand(augmented.current_state))

    @override
    def converged(self, augmented: ComparingState[State, Comparand]) -> JaxBooleanArray:
        return tree_reduce(jnp.logical_and,
                           tree_map(partial(jnp.allclose, rtol=self.rtol, atol=self.atol),
                                    self.extract_comparand(augmented.current_state),
                                    augmented.last_state),
                           jnp.asarray(True))

    @override
    def minimum_tolerances(self, augmented: ComparingState[State, Comparand]
                           ) -> tuple[JaxRealArray, JaxRealArray]:
        """The minimum tolerances that would lead to convergence now.

        Returns:
            The minimum value of atol that would lead to convergence now.
            The minimum value of rtol that would lead to convergence now.
        """
        comparand = self.extract_comparand(augmented.current_state)
        abs_last = tree_map(jnp.abs, augmented.last_state)
        delta = tree_map(jnp.abs, tree_map(jnp.subtract, comparand, augmented.last_state))
        delta_over_b = tree_map(divide_nonnegative, delta, abs_last)
        minium_atol = tree_reduce(jnp.maximum,
                                  tree_map(jnp.amax, delta), jnp.asarray(0.0))
        minium_rtol = tree_reduce(jnp.maximum,
                                  tree_map(jnp.amax, delta_over_b), jnp.asarray(0.0))
        return minium_atol, minium_rtol



from __future__ import annotations

from numbers import Number
from typing import Any, overload

import jax.numpy as jnp
import numpy as np
from jax import float0

from .annotations import (Array, BooleanNumeric, ComplexNumeric, IntegralNumeric, JaxArray,
                          JaxComplexArray, JaxRealArray, NumpyRealArray, RealNumeric)

__all__ = [
    'abs_square',
    'divide_nonnegative',
    'divide_where',
    'inverse_softplus',
    'is_scalar',
    'zero_tangent_like']


def is_scalar(x: Any) -> bool:
    return isinstance(x, Number) or isinstance(x, np.ndarray | JaxArray) and x.shape == ()


def abs_square(x: ComplexNumeric) -> JaxRealArray:
    return jnp.square(x.real) + jnp.square(x.imag)


@overload
def divide_where(dividend: RealNumeric,
                 divisor: RealNumeric | IntegralNumeric,
                 *,
                 where: BooleanNumeric | None = None,
                 otherwise: RealNumeric | None = None) -> JaxRealArray:
    ...


@overload
def divide_where(dividend: ComplexNumeric,
                 divisor: ComplexNumeric | IntegralNumeric,
                 *,
                 where: BooleanNumeric | None = None,
                 otherwise: ComplexNumeric | None = None) -> JaxComplexArray:
    ...


def divide_where(dividend: ComplexNumeric,
                 divisor: ComplexNumeric | IntegralNumeric,
                 *,
                 where: BooleanNumeric | None = None,
                 otherwise: ComplexNumeric | None = None) -> JaxComplexArray:
    """Return the quotient or a special value when a condition is false.

    Returns: `jnp.where(where, dividend / divisor, otherwise)`, but without evaluating
    `dividend / divisor` when `where` is false.  This prevents some exceptions.
    """
    if where is None:
        assert otherwise is None
        return jnp.true_divide(dividend, divisor)
    assert otherwise is not None
    dividend = jnp.where(where, dividend, 1.0)
    divisor = jnp.where(where, divisor, 1.0)
    quotient: JaxComplexArray = jnp.true_divide(dividend, divisor)
    return jnp.where(where, quotient, otherwise)


def divide_nonnegative(dividend: RealNumeric, divisor: RealNumeric) -> JaxRealArray:
    """Quotient for use with positive reals that never returns NaN.

    Returns: The quotient assuming that the dividend and divisor are nonnegative, and infinite
    whenever the divisor equals zero.
    """
    return divide_where(dividend, divisor, where=divisor > 0.0, otherwise=jnp.inf)  # noqa: PLR2004


def zero_tangent_like(value: Array) -> NumpyRealArray:
    if jnp.issubdtype(value.dtype, jnp.inexact):
        return np.zeros_like(value)
    return np.zeros_like(value, dtype=float0)


def inverse_softplus(y: RealNumeric) -> JaxRealArray:
    return jnp.where(y > 80.0,  # noqa: PLR2004
                     y,
                     jnp.log(jnp.expm1(y)))



from __future__ import annotations

from typing import Any, Generic

import jax.numpy as jnp
from jax.tree_util import tree_map, tree_reduce
from typing_extensions import override

from tjax.dataclasses import dataclass

from ..annotations import JaxBooleanArray, JaxRealArray
from .augmented import AugmentedState, State
from .combinator import Differentiand, IteratedFunctionWithCombinator
from .iterated_function import Comparand, IteratedFunction, Parameters, Trajectory

__all__ = ['ZeroIteratedFunction', 'ZeroIteratedFunctionWithCombinator']


@dataclass
class ZeroIteratedFunction(
        IteratedFunction[Parameters, State, Comparand, Trajectory,
                         AugmentedState[State]],
        Generic[Parameters, State, Comparand, Trajectory]):
    """An iterated function that converges when a part of the state is zero."""

    # Implemented methods --------------------------------------------------------------------------
    @override
    def initial_augmented(self, initial_state: State) -> AugmentedState[State]:
        return AugmentedState(current_state=initial_state,
                              iterations=0)

    @override
    def expected_state(self, theta: Parameters, state: State) -> State:
        return self.sampled_state(theta, state)

    @override
    def iterate_augmented(self,
                          new_state: State,
                          augmented: AugmentedState[State]
                          ) -> AugmentedState[State]:
        return AugmentedState(current_state=new_state,
                              iterations=augmented.iterations + 1)

    @override
    def converged(self, augmented: AugmentedState[State]) -> JaxBooleanArray:
        comparand = self.extract_comparand(augmented.current_state)

        def all_close_zero(x: Any) -> JaxBooleanArray:
            return jnp.all(jnp.abs(x) < self.atol)

        return tree_reduce(jnp.logical_and,
                           tree_map(all_close_zero, comparand),
                           jnp.asarray(True))

    @override
    def minimum_tolerances(self, augmented: AugmentedState[State]
                           ) -> tuple[JaxRealArray, JaxRealArray]:
        """The minimum tolerances that would lead to convergence now.

        Returns:
            The minimum value of atol that would lead to convergence now.
            The minimum value of rtol that would lead to convergence now.
        """
        comparand = self.extract_comparand(augmented.current_state)
        abs_comparand = tree_map(jnp.abs, comparand)
        minium_atol = tree_reduce(jnp.maximum,
                                  tree_map(jnp.amax, abs_comparand), jnp.asarray(0.0))
        return minium_atol, jnp.array(0.0)


class ZeroIteratedFunctionWithCombinator(
        IteratedFunctionWithCombinator[Parameters, State, Comparand, Differentiand, Trajectory,
                                       AugmentedState[State]],
        ZeroIteratedFunction[Parameters, State, Comparand, Trajectory],
        Generic[Parameters, State, Comparand, Differentiand, Trajectory]):
    pass



from __future__ import annotations

from functools import WRAPPER_ASSIGNMENTS

__all__: list[str] = []


abstract_jit_marker = '_abstract_jit'
abstract_custom_jvp_marker = '_abstract_custom_jvp'
all_wrapper_assignments = (*WRAPPER_ASSIGNMENTS, '__isabstractmethod__', '__override__',
                           abstract_jit_marker, abstract_custom_jvp_marker)
